//-----------------------------------------------------------------------------
//	Copyright (C) 1999, 2000 Core Technologies.
//
//	This file is part of tpasm.
//
//	tpasm is free software; you can redistribute it and/or modify
//	it under the terms of the tpasm LICENSE AGREEMENT.
//
//	tpasm is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	tpasm LICENSE AGREEMENT for more details.
//
//	You should have received a copy of the tpasm LICENSE AGREEMENT
//	along with tpasm; see the file "LICENSE.TXT".
//
//-----------------------------------------------------------------------------
//
//	Motorola MC68HC11 support Copyright (C) 1999 Cosmodog, Ltd.
//
//-----------------------------------------------------------------------------



// 68HC11 ADDRESSING MODE FORMS
//
//	inherent:				op
//	immediate8:				op	#ii
//	immediate16:			op	#jjkk
//	direct:					op	dd
//	extended:				op	hhll
//	indexed_x:				op	ff,x
//	indexed_y:				op	ff,y
//	relative:				op	rr
//	bit_direct:				op	dd,mm
//	bit_indexed_x:			op	ff,x,mm
//	bit_indexed_y:			op	ff,y,mm
//	bit_direct_relative:	op	dd,mm,rr
//	bit_indexed_x_relative:	op	ff,x,mm,rr
//	bit_indexed_y_relative:	op	ff,y,mm,rr
//
//	ii		8 bit immediate value
//	jjkk	16 bit immediate value (jj is upper 8 bits, kk is lower 8 bits)
//	dd		8 bit direct page address
//	hhll	16 bit address (hh is upper 8 bits, ll is lower 8 bits)
//	ff		8 bit unsigned offset
//	rr		8 bit signed offset
//	mm		8 bit mask
//
//	ff is an implied 0 if it is not included in indexed addressing modes, i.e.,
//		addd	x
//		addd	,x
//		addd	0,x
//	are all the same instruction (and are all valid)
//



#include	"include.h"

static SYMTABLE
	*opcodeSymbols;

// enumerated addressing modes

enum
{
	OT_INHERENT = 0,				// no operands
	OT_IMMEDIATE8,					// one byte immediate operand
	OT_IMMEDIATE16,					// two byte immediate operand
	OT_DIRECT,						// one byte direct
	OT_EXTENDED,					// two byte absolute
	OT_INDEXED_X,
	OT_INDEXED_Y,
	OT_RELATIVE,					// one byte relative offset
	OT_BIT_DIRECT,					// one byte direct, one bit mask
	OT_BIT_INDEXED_X,
	OT_BIT_INDEXED_Y,
	OT_BIT_DIRECT_RELATIVE,			// one byte direct, one bit mask
	OT_BIT_INDEXED_X_RELATIVE,
	OT_BIT_INDEXED_Y_RELATIVE,
	OT_NUM							// number of addressing modes
};


// masks for the various addressing modes

#define	M_INHERENT 						(1<<OT_INHERENT)
#define	M_IMMEDIATE8					(1<<OT_IMMEDIATE8)
#define	M_IMMEDIATE16					(1<<OT_IMMEDIATE16)
#define	M_DIRECT						(1<<OT_DIRECT)
#define	M_EXTENDED						(1<<OT_EXTENDED)
#define	M_INDEXED_X						(1<<OT_INDEXED_X)
#define	M_INDEXED_Y						(1<<OT_INDEXED_Y)
#define	M_RELATIVE						(1<<OT_RELATIVE)
#define	M_BIT_DIRECT					(1<<OT_BIT_DIRECT)
#define	M_BIT_INDEXED_X					(1<<OT_BIT_INDEXED_X)
#define	M_BIT_INDEXED_Y					(1<<OT_BIT_INDEXED_Y)
#define	M_BIT_DIRECT_RELATIVE			(1<<OT_BIT_DIRECT_RELATIVE)
#define	M_BIT_INDEXED_X_RELATIVE		(1<<OT_BIT_INDEXED_X_RELATIVE)
#define	M_BIT_INDEXED_Y_RELATIVE		(1<<OT_BIT_INDEXED_Y_RELATIVE)

typedef bool HandlerFunction(struct opcode *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);

typedef unsigned short opcode_t;

typedef struct opcode
{
	char
		*name;
	HandlerFunction
		*pseudoHandler;
	int
		typeMask;
	opcode_t						// opcodes are treated as two bytes to accomodate 'prebyte'.  if high byte is zero the opcode is actually one byte
		baseOpcode[OT_NUM];
} OPCODE;

// prototypes for the handler functions which need to be declared before we can build
// the opcode table

static bool HandleDB(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleDW(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleDS(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleDSW(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleIncbin(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);

#define	______		0					// unused opcode

static OPCODE
	Opcodes[]=
	{
		// pseudo ops

		{"db",		HandleDB,		0,																		{______,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"dc.b",	HandleDB,		0,																		{______,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"dw",		HandleDW,		0,																		{______,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"dc.w",	HandleDW,		0,																		{______,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"ds",		HandleDS,		0,																		{______,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"ds.b",	HandleDS,		0,																		{______,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"ds.w",	HandleDSW,		0,																		{______,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"incbin",	HandleIncbin,	0,																		{______,______,______,______,______,______,______,______,______,______,______,______,______,______}},

		// opcodes
//																											  inh    imm8   imm16   dir    ext   indx   indy   rel    bdir   binx   biny   bdr    bix    biy
		{"aba",		NULL,		M_INHERENT,																	{0x001b,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"abx",		NULL,		M_INHERENT,																	{0x003a,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"aby",		NULL,		M_INHERENT,																	{0x183a,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"adca",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x0089,______,0x0099,0x00b9,0x00a9,0x18a9,______,______,______,______,______,______,______}},
		{"adcb",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x00c9,______,0x00d9,0x00f9,0x00e9,0x18e9,______,______,______,______,______,______,______}},
		{"adda",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x008b,______,0x009b,0x00bb,0x00ab,0x18ab,______,______,______,______,______,______,______}},
		{"addb",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x00cb,______,0x00db,0x00fb,0x00eb,0x18eb,______,______,______,______,______,______,______}},
		{"addd",	NULL,		M_IMMEDIATE16|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,______,0x00c3,0x00d3,0x00f3,0x00e3,0x18e3,______,______,______,______,______,______,______}},
		{"anda",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x0084,______,0x0094,0x00b4,0x00a4,0x18a4,______,______,______,______,______,______,______}},
		{"andb",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x00c4,______,0x00d4,0x00f4,0x00e4,0x18e4,______,______,______,______,______,______,______}},
		{"asla",	NULL,		M_INHERENT,																	{0x0048,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"aslb",	NULL,		M_INHERENT,																	{0x0058,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"asl",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x0078,0x0068,0x1868,______,______,______,______,______,______,______}},
		{"asld",	NULL,		M_INHERENT,																	{0x0005,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"asra",	NULL,		M_INHERENT,																	{0x0047,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"asrb",	NULL,		M_INHERENT,																	{0x0057,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"asr",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x0077,0x0067,0x1867,______,______,______,______,______,______,______}},
		{"bcc",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0024,______,______,______,______,______,______}},		// same as bhs
		{"bclr",	NULL,		M_BIT_DIRECT|M_BIT_INDEXED_X|M_BIT_INDEXED_Y,								{______,______,______,______,______,______,______,______,0x0015,0x001d,0x181d,______,______,______}},
		{"bcs",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0025,______,______,______,______,______,______}},		// same as blo
		{"beq",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0027,______,______,______,______,______,______}},
		{"bge",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x002c,______,______,______,______,______,______}},
		{"bgt",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x002e,______,______,______,______,______,______}},
		{"bhi",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0022,______,______,______,______,______,______}},
		{"bhs",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0024,______,______,______,______,______,______}},		// same as bcc
		{"bita",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x0085,______,0x0095,0x00b5,0x00a5,0x18a5,______,______,______,______,______,______,______}},
		{"bitb",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x00c5,______,0x00d5,0x00f5,0x00e5,0x18e5,______,______,______,______,______,______,______}},
		{"ble",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x002f,______,______,______,______,______,______}},
		{"blo",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0025,______,______,______,______,______,______}},		// same as bcs
		{"bls",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0023,______,______,______,______,______,______}},
		{"blt",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x002d,______,______,______,______,______,______}},
		{"bmi",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x002b,______,______,______,______,______,______}},
		{"bne",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0026,______,______,______,______,______,______}},
		{"bpl",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x002a,______,______,______,______,______,______}},
		{"bra",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0020,______,______,______,______,______,______}},
		{"brclr",	NULL,		M_BIT_DIRECT_RELATIVE|M_BIT_INDEXED_X_RELATIVE|M_BIT_INDEXED_Y_RELATIVE,	{______,______,______,______,______,______,______,______,______,______,______,0x0013,0x001f,0x181f}},
		{"brn",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0021,______,______,______,______,______,______}},
		{"brset",	NULL,		M_BIT_DIRECT_RELATIVE|M_BIT_INDEXED_X_RELATIVE|M_BIT_INDEXED_Y_RELATIVE,	{______,______,______,______,______,______,______,______,______,______,______,0x0012,0x001e,0x181e}},
		{"bset",	NULL,		M_BIT_DIRECT|M_BIT_INDEXED_X|M_BIT_INDEXED_Y,								{______,______,______,______,______,______,______,______,0x0014,0x001c,0x181c,______,______,______}},
		{"bsr",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x008d,______,______,______,______,______,______}},
		{"bvc",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0028,______,______,______,______,______,______}},
		{"bvs",		NULL,		M_RELATIVE,																	{______,______,______,______,______,______,______,0x0029,______,______,______,______,______,______}},
		{"cba",		NULL,		M_INHERENT,																	{0x0011,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"clc",		NULL,		M_INHERENT,																	{0x000c,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"cli",		NULL,		M_INHERENT,																	{0x000e,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"clra",	NULL,		M_INHERENT,																	{0x004f,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"clrb",	NULL,		M_INHERENT,																	{0x005f,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"clr",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x007f,0x006f,0x186f,______,______,______,______,______,______,______}},
		{"clv",		NULL,		M_INHERENT,																	{0x000a,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"cmpa",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x0081,______,0x0091,0x00b1,0x00a1,0x18a1,______,______,______,______,______,______,______}},
		{"cmpb",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x00c1,______,0x00d1,0x00f1,0x00e1,0x18e1,______,______,______,______,______,______,______}},
		{"coma",	NULL,		M_INHERENT,																	{0x0043,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"comb",	NULL,		M_INHERENT,																	{0x0053,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"com",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x0073,0x0063,0x1863,______,______,______,______,______,______,______}},
		{"cpd",		NULL,		M_IMMEDIATE16|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,______,0x1a83,0x1a93,0x1ab3,0x1aa3,0xcda3,______,______,______,______,______,______,______}},
		{"cpx",		NULL,		M_IMMEDIATE16|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,______,0x008c,0x009c,0x00bc,0x00ac,0xcdac,______,______,______,______,______,______,______}},
		{"cpy",		NULL,		M_IMMEDIATE16|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,______,0x188c,0x189c,0x18bc,0x1aac,0x18ac,______,______,______,______,______,______,______}},
		{"daa",		NULL,		M_INHERENT,																	{0x0019,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"deca",	NULL,		M_INHERENT,																	{0x004a,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"decb",	NULL,		M_INHERENT,																	{0x005a,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"dec",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x007a,0x006a,0x186a,______,______,______,______,______,______,______}},
		{"des",		NULL,		M_INHERENT,																	{0x0034,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"dex",		NULL,		M_INHERENT,																	{0x0009,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"dey",		NULL,		M_INHERENT,																	{0x1809,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"eora",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x0088,______,0x0098,0x00b8,0x00a8,0x18a8,______,______,______,______,______,______,______}},
		{"eorb",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x00c8,______,0x00d8,0x00f8,0x00e8,0x18e8,______,______,______,______,______,______,______}},
		{"fdiv",	NULL,		M_INHERENT,																	{0x0003,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"idiv",	NULL,		M_INHERENT,																	{0x0002,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"inca",	NULL,		M_INHERENT,																	{0x004c,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"incb",	NULL,		M_INHERENT,																	{0x005c,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"inc",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x007c,0x006c,0x186c,______,______,______,______,______,______,______}},
		{"ins",		NULL,		M_INHERENT,																	{0x0031,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"inx",		NULL,		M_INHERENT,																	{0x0008,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"iny",		NULL,		M_INHERENT,																	{0x1808,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"jmp",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x007e,0x006e,0x186e,______,______,______,______,______,______,______}},
		{"jsr",		NULL,		M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,								{______,______,______,0x009d,0x00bd,0x00ad,0x18ad,______,______,______,______,______,______,______}},
		{"ldaa",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x0086,______,0x0096,0x00b6,0x00a6,0x18a6,______,______,______,______,______,______,______}},
		{"ldab",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x00c6,______,0x00d6,0x00f6,0x00e6,0x18e6,______,______,______,______,______,______,______}},
		{"ldd",		NULL,		M_IMMEDIATE16|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,______,0x00cc,0x00dc,0x00fc,0x00ec,0x18ec,______,______,______,______,______,______,______}},
		{"lds",		NULL,		M_IMMEDIATE16|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,______,0x008e,0x009e,0x00be,0x00ae,0x18ae,______,______,______,______,______,______,______}},
		{"ldx",		NULL,		M_IMMEDIATE16|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,______,0x00ce,0x00de,0x00fe,0x00ee,0xcdee,______,______,______,______,______,______,______}},
		{"ldy",		NULL,		M_IMMEDIATE16|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,______,0x18ce,0x18de,0x18fe,0x1aee,0x18ee,______,______,______,______,______,______,______}},
		{"lsla",	NULL,		M_INHERENT,																	{0x0048,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"lslb",	NULL,		M_INHERENT,																	{0x0058,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"lsl",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x0078,0x0068,0x1868,______,______,______,______,______,______,______}},
		{"lsld",	NULL,		M_INHERENT,																	{0x0005,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"lsra",	NULL,		M_INHERENT,																	{0x0044,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"lsrb",	NULL,		M_INHERENT,																	{0x0054,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"lsr",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x0074,0x0064,0x1864,______,______,______,______,______,______,______}},
		{"lsrd",	NULL,		M_INHERENT,																	{0x0004,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"mul",		NULL,		M_INHERENT,																	{0x003d,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"nega",	NULL,		M_INHERENT,																	{0x0040,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"negb",	NULL,		M_INHERENT,																	{0x0050,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"neg",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x0070,0x0060,0x1860,______,______,______,______,______,______,______}},
		{"nop",		NULL,		M_INHERENT,																	{0x0001,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"oraa",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x008a,______,0x009a,0x00ba,0x00aa,0x18aa,______,______,______,______,______,______,______}},
		{"orab",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x00ca,______,0x00da,0x00fa,0x00ea,0x18ea,______,______,______,______,______,______,______}},
		{"psha",	NULL,		M_INHERENT,																	{0x0036,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"pshb",	NULL,		M_INHERENT,																	{0x0037,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"pshx",	NULL,		M_INHERENT,																	{0x003c,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"pshy",	NULL,		M_INHERENT,																	{0x183c,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"pula",	NULL,		M_INHERENT,																	{0x0032,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"pulb",	NULL,		M_INHERENT,																	{0x0033,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"pulx",	NULL,		M_INHERENT,																	{0x0038,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"puly",	NULL,		M_INHERENT,																	{0x1838,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"rola",	NULL,		M_INHERENT,																	{0x0049,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"rolb",	NULL,		M_INHERENT,																	{0x0059,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"rol",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x0079,0x0069,0x1869,______,______,______,______,______,______,______}},
		{"rora",	NULL,		M_INHERENT,																	{0x0046,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"rorb",	NULL,		M_INHERENT,																	{0x0056,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"ror",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x0076,0x0066,0x1866,______,______,______,______,______,______,______}},
		{"rti",		NULL,		M_INHERENT,																	{0x003b,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"rts",		NULL,		M_INHERENT,																	{0x0039,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"sba",		NULL,		M_INHERENT,																	{0x0010,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"sbca",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x0082,______,0x0092,0x00b2,0x00a2,0x18a2,______,______,______,______,______,______,______}},
		{"sbcb",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x00c2,______,0x00d2,0x00f2,0x00e2,0x18e2,______,______,______,______,______,______,______}},
		{"sec",		NULL,		M_INHERENT,																	{0x000d,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"sei",		NULL,		M_INHERENT,																	{0x000f,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"sev",		NULL,		M_INHERENT,																	{0x000b,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"staa",	NULL,		M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,								{______,______,______,0x0097,0x00b7,0x00a7,0x18a7,______,______,______,______,______,______,______}},
		{"stab",	NULL,		M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,								{______,______,______,0x00d7,0x00f7,0x00e7,0x18e7,______,______,______,______,______,______,______}},
		{"std",		NULL,		M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,								{______,______,______,0x00dd,0x00fd,0x00ed,0x18ed,______,______,______,______,______,______,______}},
		{"stop",	NULL,		M_INHERENT,																	{0x00cf,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"sts",		NULL,		M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,								{______,______,______,0x009f,0x00bf,0x00af,0x18af,______,______,______,______,______,______,______}},
		{"stx",		NULL,		M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,								{______,______,______,0x00df,0x00ff,0x00ef,0xcdef,______,______,______,______,______,______,______}},
		{"sty",		NULL,		M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,								{______,______,______,0x18df,0x18ff,0x1aef,0x18ef,______,______,______,______,______,______,______}},
		{"suba",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x0080,______,0x0090,0x00b0,0x00a0,0x18a0,______,______,______,______,______,______,______}},
		{"subb",	NULL,		M_IMMEDIATE8|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,0x00c0,______,0x00d0,0x00f0,0x00e0,0x18e0,______,______,______,______,______,______,______}},
		{"subd",	NULL,		M_IMMEDIATE16|M_DIRECT|M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,					{______,______,0x0083,0x0093,0x00b3,0x00a3,0x18a3,______,______,______,______,______,______,______}},
		{"swi",		NULL,		M_INHERENT,																	{0x003f,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"tab",		NULL,		M_INHERENT,																	{0x0016,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"tap",		NULL,		M_INHERENT,																	{0x0006,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"tba",		NULL,		M_INHERENT,																	{0x0017,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"test",	NULL,		M_INHERENT,																	{0x0000,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"tpa",		NULL,		M_INHERENT,																	{0x0007,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"tsta",	NULL,		M_INHERENT,																	{0x004d,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"tstb",	NULL,		M_INHERENT,																	{0x005d,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"tst",		NULL,		M_EXTENDED|M_INDEXED_X|M_INDEXED_Y,											{______,______,______,______,0x007d,0x006d,0x186d,______,______,______,______,______,______,______}},
		{"tsx",		NULL,		M_INHERENT,																	{0x0030,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"tsy",		NULL,		M_INHERENT,																	{0x1830,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"txs",		NULL,		M_INHERENT,																	{0x0035,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"tys",		NULL,		M_INHERENT,																	{0x1835,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"wai",		NULL,		M_INHERENT,																	{0x003e,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"xgdx",	NULL,		M_INHERENT,																	{0x008f,______,______,______,______,______,______,______,______,______,______,______,______,______}},
		{"xgdy",	NULL,		M_INHERENT,																	{0x188f,______,______,______,______,______,______,______,______,______,______,______,______,______}},
	};
//																											  inh    imm8   imm16    dir    ext   indx   indy   rel    bdir   binx   biny   bdr    bix    biy

static bool GenerateByte(unsigned char value,LISTING_RECORD *listingRecord)
// output the value to the current segment
// This will return false only if a "hard" error occurs
{
	bool
		fail;
	int
		length;

	fail=false;
	if(currentSegment)
	{
		if(currentSegment->currentPC<65536)
		{
			if(!intermediatePass)				// only do the real work if necessary
			{
				length=strlen(listingRecord->listObjectString);
				if(length+2<MAXSTRING)
				{
					sprintf(&listingRecord->listObjectString[length],"%02X ",value);		// create list file output
				}
				fail=!AddBytesToSegment(currentSegment,currentSegment->currentPC,&value,1);
			}
			currentSegment->currentPC++;
		}
		else
		{
			AssemblyComplaint(NULL,true,"Attempt to write code outside address space\n");
		}
	}
	else
	{
		AssemblyComplaint(NULL,true,"Code cannot occur outside of a segment\n");
	}
	return(!fail);
}

static bool ParseOperandSeparator(char *theLine,int *lineIndex)
// Expect a comma, step over one if found
{
	if(theLine[*lineIndex]==',')			// does this look like an opcode separator?
	{
		(*lineIndex)++;						// step over it
		return(true);
	}
	return(false);
}

static bool ParseImmediatePreamble(char *theLine,int *lineIndex)
// Expect a pound sign, step over one if found
{
	if(theLine[*lineIndex]=='#')			// does this look like a '#'?
	{
		(*lineIndex)++;						// step over it
		return(true);
	}
	return(false);
}

static bool ParseIndexX(char *theLine,int *lineIndex)
// see if next thing on the line is index register X
// return true if so, false otherwise
{
	if(theLine[*lineIndex]=='X'||theLine[*lineIndex]=='x')
	{
		if(!IsLabelChar(theLine[(*lineIndex)+1]))
		{
			(*lineIndex)++;						// step over the register
			return(true);
		}
	}
	return(false);
}

static bool ParseIndexY(char *theLine,int *lineIndex)
// see if next thing on the line is index register Y
// return true if so, false otherwise
{
	if(theLine[*lineIndex]=='Y'||theLine[*lineIndex]=='y')
	{
		if(!IsLabelChar(theLine[(*lineIndex)+1]))
		{
			(*lineIndex)++;						// step over the register
			return(true);
		}
	}
	return(false);
}

// opcode and processor-specific pseudo opcode handling for 6805

// pseudo op handling

static bool HandleDB(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// declaring bytes of data
{
	int
		value;
	bool
		unresolved;
	char
		outputString[MAXSTRING];
	int
		stringLength;
	bool
		done,
		fail;
	int
		i;

	fail=false;
	done=false;
	while(!done&&!fail)
	{
		if(ParseQuotedString(theLine,lineIndex,'"','"',outputString,&stringLength))
		{
			i=0;
			while(!fail&&i<stringLength)
			{
				fail=!GenerateByte(outputString[i],listingRecord);
				i++;
			}
		}
		else if(ParseExpression(theLine,lineIndex,&value,&unresolved))
		{
			CheckByteRange(value,true,true);
			fail=!GenerateByte(value&0xFF,listingRecord);
		}
		else
		{
			ReportBadOperands();
			done=true;
		}
		if(!done&&!fail)				// look for separator, or comment
		{
			if(ParseComment(theLine,lineIndex))
			{
				done=true;
			}
			else if(ParseOperandSeparator(theLine,lineIndex))
			{
			}
			else
			{
				ReportBadOperands();
				done=true;
			}
		}
	}
	return(!fail);
}

static bool HandleDW(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// declaring words of data in big endian
{
	int
		value;
	bool
		unresolved;
	char
		outputString[MAXSTRING];
	int
		stringLength;
	bool
		done,
		fail;
	int
		i;

	fail=false;
	done=false;
	while(!done&&!fail)
	{
		if(ParseQuotedString(theLine,lineIndex,'"','"',outputString,&stringLength))
		{
			i=0;
			while(!fail&&i<stringLength)
			{
				if(GenerateByte(0,listingRecord))
				{
					fail=!GenerateByte(outputString[i],listingRecord);
				}
				else
				{
					fail=true;
				}
				i++;
			}
		}
		else if(ParseExpression(theLine,lineIndex,&value,&unresolved))
		{
			CheckWordRange(value,true,true);
			if(GenerateByte(value>>8,listingRecord))
			{
				fail=!GenerateByte(value&0xFF,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			ReportBadOperands();
			done=true;
		}
		if(!done&&!fail)				// look for separator, or comment
		{
			if(ParseComment(theLine,lineIndex))
			{
				done=true;
			}
			else if(ParseOperandSeparator(theLine,lineIndex))
			{
			}
			else
			{
				ReportBadOperands();
				done=true;
			}
		}
	}
	return(!fail);
}

static bool HandleDS(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// move the PC forward by the given amount
{
	int
		value;
	bool
		unresolved;

	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			if(currentSegment)
			{
				if(!unresolved)
				{
					currentSegment->currentPC+=value;
				}
			}
			else
			{
				AssemblyComplaint(NULL,true,"'%s' outside of segment\n",opcode->name);
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(true);
}

static bool HandleDSW(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// move the PC forward by the given amount times two (do as if skipping this many words)
{
	int
		value;
	bool
		unresolved;

	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			value*=2;
			if(currentSegment)
			{
				if(!unresolved)
				{
					currentSegment->currentPC+=value;
				}
			}
			else
			{
				AssemblyComplaint(NULL,true,"'%s' outside of segment\n",opcode->name);
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(true);
}

static bool HandleIncbin(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// include the data of the given file into the input stream
// If there is a problem, report it and return false
{
	bool
		fail;
	char
		fileName[MAXSTRING];
	int
		stringLength;
	FILE
		*theFile;
	char
		inputDataBuffer[1024];
	int
		bytesRead;
	int
		i;

	fail=false;

	if(ParseQuotedString(theLine,lineIndex,'"','"',fileName,&stringLength))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			if((theFile=fopen(fileName,"rb")))
			{
				while(!fail&&((bytesRead=fread(inputDataBuffer,1,1024,theFile))>0))
				{
					for(i=0;i<bytesRead;i++)
					{
						fail=!GenerateByte(inputDataBuffer[i],listingRecord);	// ### this is really inefficient
					}
				}
				fclose(theFile);
			}
			else
			{
				AssemblyComplaint(NULL,true,"Failed to open file '%s'. %s\n",fileName,strerror(errno));
			}
		}
		else
		{
			AssemblyComplaint(NULL,true,"Missing or incorrectly formatted file name\n");
		}
	}
	else
	{
		AssemblyComplaint(NULL,true,"Missing or incorrectly formatted file name\n");
	}
	return(!fail);
}

// opcode handling

enum
{
	POT_INDEX_X,
	POT_INDEX_Y,
	POT_IMMEDIATE,
	POT_VALUE,
};


static bool ParseOperandElement(char *theLine,int *lineIndex,int *type,int *value,bool *unresolved)
// Try to parse an element of the operand and determine its type
// return true if the parsing succeeds
{
	if(ParseImmediatePreamble(theLine,lineIndex))		// an immediate operand?
	{
		if(ParseExpression(theLine,lineIndex,value,unresolved))
		{
			*type=POT_IMMEDIATE;
			return(true);
		}
	}
	else if(ParseIndexX(theLine,lineIndex))
	{
		*type=POT_INDEX_X;
		return(true);
	}
	else if(ParseIndexY(theLine,lineIndex))
	{
		*type=POT_INDEX_Y;
		return(true);
	}
	else if(ParseExpression(theLine,lineIndex,value,unresolved))
	{
		*type=POT_VALUE;
		return(true);
	}
	return(false);
}


static bool WriteOpcode(opcode_t opcode, LISTING_RECORD *listingRecord)
{
	bool fail = false;
	if(opcode & 0xff00)		// if it has a pre-byte (opcode > 0xff) put out both bytes
	{
		fail = !GenerateByte((opcode>>8)&0xff,listingRecord);
	}
	if(!fail)
	{
		fail = !GenerateByte(opcode&0xff,listingRecord);
	}
	return(!fail);
}


static bool HandleImmediate(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// an immediate addressing mode was located
{
	bool fail=false;

	if(theOpcode->typeMask&M_IMMEDIATE8)			// 8-bit immediate value
	{
		CheckByteRange(value,true,true);
		fail = !WriteOpcode(theOpcode->baseOpcode[OT_IMMEDIATE8],listingRecord);
		if(!fail)
		{
			fail=!GenerateByte(value,listingRecord);
		}
	}
	else if(theOpcode->typeMask&M_IMMEDIATE16)		// 16-bit immediate value
	{
		CheckWordRange(value,true,true);
		fail = !WriteOpcode(theOpcode->baseOpcode[OT_IMMEDIATE16],listingRecord);
		if(!fail)
		{
			fail = !GenerateByte((value>>8)&0xff,listingRecord);
			if(!fail)
			{
				fail = !GenerateByte(value&0xff,listingRecord);
			}
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleRelative(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// a relative operand has been parsed
{
	bool
		fail;
	int
		offset;

	fail = !WriteOpcode(theOpcode->baseOpcode[OT_RELATIVE],listingRecord);
	if(!fail)
	{
		if(unresolved)
		{
			offset=0;
		}
		else
		{
			offset=value-(currentSegment->currentPC+currentSegment->codeGenOffset)-1;
			Check8RelativeRange(offset,true,true);
		}
		fail=!GenerateByte(offset,listingRecord);
	}
	return(!fail);
}

static bool HandleDirect(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// deal with direct mode output only
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_DIRECT)
	{
		CheckUnsignedByteRange(value,true,true);
		fail = !WriteOpcode(theOpcode->baseOpcode[OT_DIRECT],listingRecord);
		if(!fail)
		{
			fail=!GenerateByte(value,listingRecord);
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleExtended(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// deal with extended mode output only
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_EXTENDED)
	{
		CheckUnsignedWordRange(value,true,true);
		fail = !WriteOpcode(theOpcode->baseOpcode[OT_EXTENDED],listingRecord);
		if(!fail)
		{
			fail = !GenerateByte(value>>8,listingRecord);
			if(!fail)
			{
				fail=!GenerateByte(value&0xFF,listingRecord);
			}
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleDirectOrExtended(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// a direct or extended address has been parsed
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_DIRECT)
	{
		if(((value>=0)&&(value<256))||!(theOpcode->typeMask&M_EXTENDED))
		{
			fail=!HandleDirect(theOpcode,value,unresolved,listingRecord);
		}
		else
		{
			fail=!HandleExtended(theOpcode,value,unresolved,listingRecord);
		}
	}
	else
	{
		fail=!HandleExtended(theOpcode,value,unresolved,listingRecord);
	}
	return(!fail);
}

static bool HandleSingleAddress(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// a lone address has been parsed
// This means it is either a relative address (rr), direct (dd), or extended (hhll)
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_RELATIVE)
	{
		fail=!HandleRelative(theOpcode,value,unresolved,listingRecord);
	}
	else if(theOpcode->typeMask&(M_DIRECT|M_EXTENDED))
	{
		fail=!HandleDirectOrExtended(theOpcode,value,unresolved,listingRecord);
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleIndexedX(OPCODE *theOpcode,char *theLine,int *lineIndex,int value1,bool unresolved,LISTING_RECORD *listingRecord)
// indexed x modes:
//  indexed_x:				op	ff,x
//  bit_indexed_x:			op	ff,x,mm
//  bit_indexed_x_relative:	op	ff,x,mm,rr
{
	bool fail = false;
	int value2;
	int value3;
	int offset;
	bool unresolved2;
	bool unresolved3;
	int elementType;

	if(ParseOperandSeparator(theLine,lineIndex))		// look for a second separator
	{
		if(ParseOperandElement(theLine,lineIndex,&elementType,&value2,&unresolved2) && (elementType == POT_VALUE) )
		{
			if(ParseOperandSeparator(theLine,lineIndex))	// look for a third separator
			{
				if(ParseOperandElement(theLine,lineIndex,&elementType,&value3,&unresolved3) && (elementType == POT_VALUE) &&
					ParseComment(theLine,lineIndex) && theOpcode->typeMask&M_BIT_INDEXED_X_RELATIVE )
				{
					fail = !WriteOpcode(theOpcode->baseOpcode[OT_BIT_INDEXED_X_RELATIVE],listingRecord);
					if(!fail)
					{
						CheckUnsignedByteRange(value1,true,true);		// bit indexed x relative
						fail=!GenerateByte(value1,listingRecord);
						if(!fail)
						{
							CheckUnsignedByteRange(value2,true,true);
							fail=!GenerateByte(value2,listingRecord);
						}
							if(!fail)
							{
								offset=value3-(currentSegment->currentPC+currentSegment->codeGenOffset)-1;
								Check8RelativeRange(offset,true,true);
								fail=!GenerateByte(offset,listingRecord);
							}
					}
				}
				else
				{
					ReportBadOperands();
				}
			}
			else if(ParseComment(theLine,lineIndex) && theOpcode->typeMask&M_BIT_INDEXED_X)			// bit indexed x
			{
				CheckUnsignedByteRange(value1,true,true);
				CheckUnsignedByteRange(value2,true,true);
				fail = !WriteOpcode(theOpcode->baseOpcode[OT_BIT_INDEXED_X],listingRecord);
				if(!fail)
				{
					fail=!GenerateByte(value1,listingRecord);
					if(!fail)
					{
						fail=!GenerateByte(value2,listingRecord);
					}
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else if(ParseComment(theLine,lineIndex))
	{
		if(theOpcode->typeMask&M_INDEXED_X)
		{
			CheckUnsignedByteRange(value1,true,true);
			fail = !WriteOpcode(theOpcode->baseOpcode[OT_INDEXED_X],listingRecord);
			if(!fail)
			{
				fail=!GenerateByte(value1,listingRecord);
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}


static bool HandleIndexedY(OPCODE *theOpcode,char *theLine,int *lineIndex,int value1,bool unresolved,LISTING_RECORD *listingRecord)
// indexed y modes:
//  indexed_y:				op	ff,y
//  bit_indexed_y:			op	ff,y,mm
//  bit_indexed_y_relative:	op	ff,y,mm,rr
{
	bool fail = false;
	int value2;
	int value3;
	int offset;
	bool unresolved2;
	bool unresolved3;
	int elementType;

	if(ParseOperandSeparator(theLine,lineIndex))		// look for a second separator
	{
		if(ParseOperandElement(theLine,lineIndex,&elementType,&value2,&unresolved2) && (elementType == POT_VALUE) )
		{
			if(ParseOperandSeparator(theLine,lineIndex))	// look for a third separator
			{
				if(ParseOperandElement(theLine,lineIndex,&elementType,&value3,&unresolved3) && (elementType == POT_VALUE) &&
					ParseComment(theLine,lineIndex) && theOpcode->typeMask&M_BIT_INDEXED_Y_RELATIVE )
				{
					fail = !WriteOpcode(theOpcode->baseOpcode[OT_BIT_INDEXED_Y_RELATIVE],listingRecord);
					if(!fail)
					{
						CheckUnsignedByteRange(value1,true,true);		// bit indexed y relative
						fail=!GenerateByte(value1,listingRecord);
						if(!fail)
						{
							CheckUnsignedByteRange(value2,true,true);
							fail=!GenerateByte(value2,listingRecord);
						}
							if(!fail)
							{
								offset=value3-(currentSegment->currentPC+currentSegment->codeGenOffset)-1;
								Check8RelativeRange(offset,true,true);
								fail=!GenerateByte(offset,listingRecord);
							}
					}
				}
				else
				{
					ReportBadOperands();
				}
			}
			else if(ParseComment(theLine,lineIndex) && theOpcode->typeMask&M_BIT_INDEXED_Y)			// bit indexed y
			{
				CheckUnsignedByteRange(value1,true,true);
				CheckUnsignedByteRange(value2,true,true);
				fail = !WriteOpcode(theOpcode->baseOpcode[OT_BIT_INDEXED_Y],listingRecord);
				if(!fail)
				{
					fail=!GenerateByte(value1,listingRecord);
					if(!fail)
					{
						fail=!GenerateByte(value2,listingRecord);
					}
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else if(ParseComment(theLine,lineIndex))
	{
		if(theOpcode->typeMask&M_INDEXED_Y)
		{
			fail = !WriteOpcode(theOpcode->baseOpcode[OT_INDEXED_Y],listingRecord);
			if(!fail)
			{
				CheckUnsignedByteRange(value1,true,true);
				fail=!GenerateByte(value1,listingRecord);
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}


static bool HandleBitDirect(OPCODE *theOpcode,char *theLine,int *lineIndex,int value1, int value2,LISTING_RECORD *listingRecord)
{
	bool fail = false;
	int value3;
	bool unresolved3;
	int elementType;
	int offset;

	if(ParseOperandSeparator(theLine,lineIndex))		// separator indicates a third value
	{
		if(ParseOperandElement(theLine,lineIndex,&elementType,&value3,&unresolved3) && elementType==POT_VALUE)
		{
			if(ParseComment(theLine,lineIndex) && theOpcode->typeMask&M_BIT_DIRECT_RELATIVE)
			{
				fail = !WriteOpcode(theOpcode->baseOpcode[OT_BIT_DIRECT_RELATIVE],listingRecord);
				if(!fail)
				{
					CheckUnsignedByteRange(value1,true,true);
					fail=!GenerateByte(value1,listingRecord);
					if(!fail)
					{
						CheckUnsignedByteRange(value2,true,true);
						fail=!GenerateByte(value2,listingRecord);
						if(!fail)
						{
							offset=value3-(currentSegment->currentPC+currentSegment->codeGenOffset)-1;
							Check8RelativeRange(offset,true,true);
							fail=!GenerateByte(offset,listingRecord);
						}
					}
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else if(ParseComment(theLine,lineIndex) && theOpcode->typeMask&M_BIT_DIRECT)
	{
		fail = !WriteOpcode(theOpcode->baseOpcode[OT_BIT_DIRECT],listingRecord);
		if(!fail)
		{
			CheckUnsignedByteRange(value1,true,true);
			fail=!GenerateByte(value1,listingRecord);
			if(!fail)
			{
				CheckUnsignedByteRange(value2,true,true);
				fail=!GenerateByte(value2,listingRecord);
			}
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);

}


static bool HandleFirstValue(OPCODE *theOpcode,char *theLine,int *lineIndex,int value1,bool unresolved1,LISTING_RECORD *listingRecord)
// a non-immediate value was parsed, so see what else we can find
{
	bool
		fail;
	int
		elementType;
	int
		value2;
	bool
		unresolved2;

	fail=false;
	if(ParseOperandSeparator(theLine,lineIndex))		// make sure the next thing separates operands
	{
		if(ParseOperandElement(theLine,lineIndex,&elementType,&value2,&unresolved2))
		{
			switch(elementType)
			{
				case POT_INDEX_X:							// second operand was X
					return(HandleIndexedX(theOpcode,theLine,lineIndex,value1,unresolved1,listingRecord));
					break;

				case POT_INDEX_Y:							// second operand was Y
					return(HandleIndexedY(theOpcode,theLine,lineIndex,value1,unresolved1,listingRecord));
					break;

				case POT_VALUE:
					return(HandleBitDirect(theOpcode,theLine,lineIndex,value1,value2,listingRecord));
					break;

				default:
					ReportBadOperands();
					break;
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}


static bool HandleOpcode(void *theOpcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// look at the type of opcode available, parse operands as allowed
// return false only if there was a 'hard' error
//
{
	OPCODE
		*actualOpcode;
	bool
		fail;
	int
		elementType;
	int
		value;
	bool
		unresolved;
	bool
		indexedMode;

	fail=false;
	actualOpcode=(OPCODE *)theOpcode;
	if(actualOpcode->pseudoHandler)
	{
		fail=!actualOpcode->pseudoHandler(actualOpcode,theLine,lineIndex,listingRecord);
	}
	else
	{
		if(!ParseComment(theLine,lineIndex))
		{
			indexedMode = ParseOperandSeparator(theLine,lineIndex);		// if true mode must be indexed
			if(ParseOperandElement(theLine,lineIndex,&elementType,&value,&unresolved))
			{
				switch(elementType)
				{
					case POT_IMMEDIATE:
						if(!indexedMode && ParseComment(theLine,lineIndex))
						{
							return(HandleImmediate(actualOpcode,value,unresolved,listingRecord));
						}
						else
						{
							ReportBadOperands();
						}
						break;

					case POT_INDEX_X:
						value = 0;
						return(HandleIndexedX(actualOpcode,theLine,lineIndex,value,unresolved,listingRecord));
						break;

					case POT_INDEX_Y:
						value = 0;
						return(HandleIndexedY(actualOpcode,theLine,lineIndex,value,unresolved,listingRecord));
						break;

					case POT_VALUE:
						if(!indexedMode && ParseComment(theLine,lineIndex))			// if it's followed by comment/EOL, then its a single address
						{
							return(HandleSingleAddress(actualOpcode,value,unresolved,listingRecord));
						}
						else
						{
							return(HandleFirstValue(actualOpcode,theLine,lineIndex,value,unresolved,listingRecord));
						}
						break;
				}
			}
			else
			{
				ReportBadOperands();
			}

		}
		else
		{
			if(actualOpcode->typeMask&M_INHERENT)
			{
				fail = !WriteOpcode(actualOpcode->baseOpcode[OT_INHERENT],listingRecord);
			}
			else
			{
				ReportBadOperands();
			}
		}
	}
	return(!fail);
}

static void *MatchOpcode(char *theOpcode)
// match opcodes for this processor, return NULL if none matched
{
	return(STFindDataForName(opcodeSymbols,theOpcode));
}

static bool SelectProcessor(PROCESSOR *theProcessor)
// A processor in this family is being selected to assemble with
{
	return(true);
}

static void DeselectProcessor(PROCESSOR *theProcessor)
// A processor in this family is being deselected
{
}

static void UnInitFamily()
// undo what InitFamily did
{
	STDisposeSymbolTable(opcodeSymbols);
}

static bool InitFamily()
// initialize symbol table
{
	unsigned int
		i;
	bool
		fail;

	fail=false;
	if((opcodeSymbols=STNewSymbolTable()))
	{
		for(i=0;!fail&&(i<sizeof(Opcodes)/sizeof(OPCODE));i++)
		{
			if(!STAddEntryAtEnd(opcodeSymbols,Opcodes[i].name,&Opcodes[i]))
			{
				fail=true;
			}
		}
		if(!fail)
		{
			return(true);
		}
		STDisposeSymbolTable(opcodeSymbols);
	}
	return(false);
}

// processors handled here (the constuctors for these variables link them to the global
// list of processors that the assembler knows how to handle)

static PROCESSOR_FAMILY
	processorFamily("Motorola 68hc11",InitFamily,UnInitFamily,SelectProcessor,DeselectProcessor,MatchOpcode,HandleOpcode);

static PROCESSOR
	processor0(&processorFamily,"68hc11",NULL);
