//	Copyright (C) 1999, 2000 Core Technologies.
//
//	This file is part of tpasm.
//
//	tpasm is free software; you can redistribute it and/or modify
//	it under the terms of the tpasm LICENSE AGREEMENT.
//
//	tpasm is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	tpasm LICENSE AGREEMENT for more details.
//
//	You should have received a copy of the tpasm LICENSE AGREEMENT
//	along with tpasm; see the file "LICENSE.TXT".


// Generate code for all pic families: 16C5X, 16CXX, and 17CXX

#include	"include.h"

static SYMTABLE
	*opcodeSymbolsPseudo,			// pseudo ops shared between all processors
	*opcodeSymbols16C5X,			// symbols for the various PIC processor families
	*opcodeSymbols16CXX,
	*opcodeSymbols17CXX;

enum
{
	PIC_16C5X,						// families
	PIC_16CXX,
	PIC_17CXX
};

typedef struct
{
	int
		family;						// tells which family this processor belongs to
	unsigned int
		coreMask;					// tells how many bits of each core location are available
	unsigned int
		bankSelBits;				// tells how many bits of bank select are needed by this processor
	unsigned int
		bankISelBits;				// tells how many bits of indirect bank select to use
	unsigned int
		ROMSize;					// tells how many locations are in ROM
	unsigned int
		configAddress;				// tells where configuration information is to be written
	unsigned int
		IDLocAddress;				// address in which to store ID data for this processor (0 if invalid)
} PIC_PROCESSOR;

static PIC_PROCESSOR
	*currentProcessor;				// points at the record for the currently selected processor

#define	MAX_BAD_RAM	0x1000			// MAXRAM pseudo-op must have argument less than this

static bool
	testBadRAM;						// tells if bad ram should be tested for
static int
	maxRAM;							// the value of maxRAM that was given
static unsigned char
	badRAMMap[MAX_BAD_RAM>>3];		// bitmap of bad RAM locations (tested when maxram is specified)

enum
{
	OT_IMPLICIT,					// no operands
	OT_LITERAL8,					// one byte literal operand
	OT_CALL8,						// 8 bit literal operand (errors generated if out of range)
	OT_CALL11,						// 11 bit operand
	OT_GOTO9,						// 9 bit operand
	OT_GOTO11,						// 11 bit operand
	OT_GOTO13,						// 13 bit operand
	OT_REGISTER3,					// 3 bits of register
	OT_REGISTER5,					// 5 bits of register
	OT_REGISTER5_DEST,				// 5 bits of register and a destination
	OT_REGISTER5_BIT,				// 5 bits of register and bit index
	OT_REGISTER7,					// 7 bits of register
	OT_REGISTER7_DEST,				// 7 bits of register and a destination
	OT_REGISTER7_BIT,				// 7 bits of register and 3 bit index
	OT_REGISTER8,					// 8 bits of register
	OT_REGISTER8_DEST,				// 8 bits of register and a destination
	OT_REGISTER8_PORT,				// 8 bits of register and 5 bits of port
	OT_PORT_REGISTER8,				// 5 bits of port and 8 bits of register
	OT_HIGHLOW_INCREMENT_REGISTER8,	// 1 bit of high/low byte, 1 bit of increment and 8 bits of register
	OT_HIGHLOW_REGISTER8,			// 1 bit of high/low byte and 8 bits of register
	OT_REGISTER8_BIT,				// 8 bits of register and 3 bit bit index
};

typedef bool HandlerFunction(struct opcode *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);

typedef struct opcode
{
	char
		*name;
	HandlerFunction
		*pseudoHandler;
	int
		type;
	int
		baseOpcode;
} OPCODE;

// prototypes for the handler functions which need to be declared before we can build
// the opcode tables

static bool HandleDB(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleDW(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleDT(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleDS(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleConfig(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleIDLocs(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleMaxRAM(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleBadRAM(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleBankSel(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleBankISel(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandlePageSel(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);

static OPCODE
	OpcodesPseudo[]=
	{
		{"db",		HandleDB,		0,						0x0000},
		{"dw",		HandleDW,		0,						0x0000},
		{"data",	HandleDW,		0,						0x0000},		// same as dw
		{"dt",		HandleDT,		0,						0x0000},
		{"ds",		HandleDS,		0,						0x0000},
		{"__config",HandleConfig,	0,						0x0000},
		{"__idlocs",HandleIDLocs,	0,						0x0000},
		{"__maxram",HandleMaxRAM,	0,						0x0000},
		{"__badram",HandleBadRAM,	0,						0x0000},
		{"banksel",	HandleBankSel,	0,						0x0000},
		{"bankisel",HandleBankISel,	0,						0x0000},
		{"pagesel",	HandlePageSel,	0,						0x0000},
	};

static OPCODE
	Opcodes16C5X[]=
	{
		{"andlw",	NULL,	OT_LITERAL8,					0x0E00},
		{"call",	NULL,	OT_CALL8,						0x0900},
		{"clrwdt",	NULL,	OT_IMPLICIT,					0x0004},
		{"goto",	NULL,	OT_GOTO9,						0x0A00},
		{"iorlw",	NULL,	OT_LITERAL8,					0x0D00},
		{"movlw",	NULL,	OT_LITERAL8,					0x0C00},
		{"option",	NULL,	OT_IMPLICIT,					0x0002},
		{"retlw",	NULL,	OT_LITERAL8,					0x0800},
		{"sleep",	NULL,	OT_IMPLICIT,					0x0003},
		{"tris",	NULL,	OT_REGISTER3,					0x0000},
		{"xorlw",	NULL,	OT_LITERAL8,					0x0F00},
		{"addwf",	NULL,	OT_REGISTER5_DEST,				0x01C0},
		{"andwf",	NULL,	OT_REGISTER5_DEST,				0x0140},
		{"clrf",	NULL,	OT_REGISTER5,					0x0060},
		{"clrw",	NULL,	OT_IMPLICIT,					0x0040},
		{"comf",	NULL,	OT_REGISTER5_DEST,				0x0240},
		{"decf",	NULL,	OT_REGISTER5_DEST,				0x00C0},
		{"decfsz",	NULL,	OT_REGISTER5_DEST,				0x02C0},
		{"incf",	NULL,	OT_REGISTER5_DEST,				0x0280},
		{"incfsz",	NULL,	OT_REGISTER5_DEST,				0x03C0},
		{"iorwf",	NULL,	OT_REGISTER5_DEST,				0x0100},
		{"movf",	NULL,	OT_REGISTER5_DEST,				0x0200},
		{"movwf",	NULL,	OT_REGISTER5,					0x0020},
		{"nop",		NULL,	OT_IMPLICIT,					0x0000},
		{"rlf",		NULL,	OT_REGISTER5_DEST,				0x0340},
		{"rrf",		NULL,	OT_REGISTER5_DEST,				0x0300},
		{"subwf",	NULL,	OT_REGISTER5_DEST,				0x0080},
		{"swapf",	NULL,	OT_REGISTER5_DEST,				0x0380},
		{"xorwf",	NULL,	OT_REGISTER5_DEST,				0x0180},
		{"bcf",		NULL,	OT_REGISTER5_BIT,				0x0400},
		{"bsf",		NULL,	OT_REGISTER5_BIT,				0x0500},
		{"btfsc",	NULL,	OT_REGISTER5_BIT,				0x0600},
		{"btfss",	NULL,	OT_REGISTER5_BIT,				0x0700},
	};

static OPCODE
	Opcodes16CXX[]=
	{
		{"addlw",	NULL,	OT_LITERAL8,					0x3E00},
		{"andlw",	NULL,	OT_LITERAL8,					0x3900},
		{"call",	NULL,	OT_CALL11,						0x2000},
		{"clrwdt",	NULL,	OT_IMPLICIT,					0x0064},
		{"goto",	NULL,	OT_GOTO11,						0x2800},
		{"iorlw",	NULL,	OT_LITERAL8,					0x3800},
		{"movlw",	NULL,	OT_LITERAL8,					0x3000},
		{"option",	NULL,	OT_IMPLICIT,					0x0062},
		{"retfie",	NULL,	OT_IMPLICIT,					0x0009},
		{"retlw",	NULL,	OT_LITERAL8,					0x3400},
		{"return",	NULL,	OT_IMPLICIT,					0x0008},
		{"sleep",	NULL,	OT_IMPLICIT,					0x0063},
		{"tris",	NULL,	OT_REGISTER3,					0x0060},
		{"xorlw",	NULL,	OT_LITERAL8,					0x3A00},
		{"addwf",	NULL,	OT_REGISTER7_DEST,				0x0700},
		{"andwf",	NULL,	OT_REGISTER7_DEST,				0x0500},
		{"clrf",	NULL,	OT_REGISTER7,					0x0180},
		{"clrw",	NULL,	OT_IMPLICIT,					0x0100},
		{"comf",	NULL,	OT_REGISTER7_DEST,				0x0900},
		{"decf",	NULL,	OT_REGISTER7_DEST,				0x0300},
		{"decfsz",	NULL,	OT_REGISTER7_DEST,				0x0B00},
		{"incf",	NULL,	OT_REGISTER7_DEST,				0x0A00},
		{"incfsz",	NULL,	OT_REGISTER7_DEST,				0x0F00},
		{"iorwf",	NULL,	OT_REGISTER7_DEST,				0x0400},
		{"movf",	NULL,	OT_REGISTER7_DEST,				0x0800},
		{"movwf",	NULL,	OT_REGISTER7,					0x0080},
		{"nop",		NULL,	OT_IMPLICIT,					0x0000},
		{"rlf",		NULL,	OT_REGISTER7_DEST,				0x0D00},
		{"rrf",		NULL,	OT_REGISTER7_DEST,				0x0C00},
		{"sublw",	NULL,	OT_LITERAL8,					0x3C00},
		{"subwf",	NULL,	OT_REGISTER7_DEST,				0x0200},
		{"swapf",	NULL,	OT_REGISTER7_DEST,				0x0E00},
		{"xorwf",	NULL,	OT_REGISTER7_DEST,				0x0600},
		{"bcf",		NULL,	OT_REGISTER7_BIT,				0x1000},
		{"bsf",		NULL,	OT_REGISTER7_BIT,				0x1400},
		{"btfsc",	NULL,	OT_REGISTER7_BIT,				0x1800},
		{"btfss",	NULL,	OT_REGISTER7_BIT,				0x1C00},
	};

static OPCODE
	Opcodes17CXX[]=
	{
		{"addlw",	NULL,	OT_LITERAL8,					0xB100},
		{"addwf",	NULL,	OT_REGISTER8_DEST,				0x0E00},
		{"addwfc",	NULL,	OT_REGISTER8_DEST,				0x1000},
		{"andlw",	NULL,	OT_LITERAL8,					0xB500},
		{"andwf",	NULL,	OT_REGISTER8_DEST,				0x0A00},
		{"bcf",		NULL,	OT_REGISTER8_BIT,				0x8800},
		{"bsf",		NULL,	OT_REGISTER8_BIT,				0x8000},
		{"btfsc",	NULL,	OT_REGISTER8_BIT,				0x9800},
		{"btfss",	NULL,	OT_REGISTER8_BIT,				0x9000},
		{"btg",		NULL,	OT_REGISTER8_BIT,				0x3800},
		{"call",	NULL,	OT_GOTO13,						0xE000},
		{"clrf",	NULL,	OT_REGISTER8_DEST,				0x2800},
		{"clrwdt",	NULL,	OT_IMPLICIT,					0x0004},
		{"comf",	NULL,	OT_REGISTER8_DEST,				0x1200},
		{"cpfseq",	NULL,	OT_REGISTER8,					0x3100},
		{"cpfsgt",	NULL,	OT_REGISTER8,					0x3200},
		{"cpfslt",	NULL,	OT_REGISTER8,					0x3000},
		{"daw",		NULL,	OT_REGISTER8_DEST,				0x2E00},
		{"dcfsnz",	NULL,	OT_REGISTER8_DEST,				0x2600},
		{"decf",	NULL,	OT_REGISTER8_DEST,				0x0600},
		{"decfsz",	NULL,	OT_REGISTER8_DEST,				0x1600},
		{"goto",	NULL,	OT_GOTO13,						0xC000},
		{"incf",	NULL,	OT_REGISTER8_DEST,				0x1400},
		{"incfsz",	NULL,	OT_REGISTER8_DEST,				0x1E00},
		{"infsnz",	NULL,	OT_REGISTER8_DEST,				0x2400},
		{"iorlw",	NULL,	OT_LITERAL8,					0xB300},
		{"iorwf",	NULL,	OT_REGISTER8_DEST,				0x0800},
		{"lcall",	NULL,	OT_LITERAL8,					0xB700},
		{"movfp",	NULL,	OT_REGISTER8_PORT,				0x6000},
		{"movlb",	NULL,	OT_LITERAL8,					0xB800},
		{"movlr",	NULL,	OT_LITERAL8,					0xBA00},
		{"movlw",	NULL,	OT_LITERAL8,					0xB000},
		{"movpf",	NULL,	OT_PORT_REGISTER8,				0x4000},
		{"movwf",	NULL,	OT_REGISTER8,					0x0100},
		{"mullw",	NULL,	OT_LITERAL8,					0xBC00},
		{"mulwf",	NULL,	OT_REGISTER8,					0x3400},
		{"negw",	NULL,	OT_REGISTER8_DEST,				0x2C00},
		{"nop",		NULL,	OT_IMPLICIT,					0x0000},
		{"ret",		NULL,	OT_IMPLICIT,					0x0002},
		{"retfie",	NULL,	OT_IMPLICIT,					0x0005},
		{"retlw",	NULL,	OT_LITERAL8,					0xB600},
		{"return",	NULL,	OT_IMPLICIT,					0x0002},
		{"rlcf",	NULL,	OT_REGISTER8_DEST,				0x1A00},
		{"rlncf",	NULL,	OT_REGISTER8_DEST,				0x2200},
		{"rrcf",	NULL,	OT_REGISTER8_DEST,				0x1800},
		{"rrncf",	NULL,	OT_REGISTER8_DEST,				0x2000},
		{"setf",	NULL,	OT_REGISTER8_DEST,				0x2A00},
		{"sleep",	NULL,	OT_IMPLICIT,					0x0003},
		{"sublw",	NULL,	OT_LITERAL8,					0xB200},
		{"subwf",	NULL,	OT_REGISTER8_DEST,				0x0400},
		{"subwfb",	NULL,	OT_REGISTER8_DEST,				0x0200},
		{"swapf",	NULL,	OT_REGISTER8_DEST,				0x1C00},
		{"tablrd",	NULL,	OT_HIGHLOW_INCREMENT_REGISTER8,	0xA800},
		{"tablwt",	NULL,	OT_HIGHLOW_INCREMENT_REGISTER8,	0xAC00},
		{"tlrd",	NULL,	OT_HIGHLOW_REGISTER8,			0xA000},
		{"tlwt",	NULL,	OT_HIGHLOW_REGISTER8,			0xA400},
		{"tstfsz",	NULL,	OT_REGISTER8,					0x3300},
		{"xorlw",	NULL,	OT_LITERAL8,					0xB400},
		{"xorwf",	NULL,	OT_REGISTER8_DEST,				0x0C00},
	};

static bool GenerateWordsAtAddress(int address,int numWords,int *wordValues,LISTING_RECORD *listingRecord)
// force word values out to memory starting at the given address
{
	unsigned char
		outputBytes[2];
	int
		temp;
	int
		length;
	int
		i;
	bool
		fail;

	fail=false;
	if(currentSegment)
	{
		if(!intermediatePass)				// only do the real work if necessary
		{
			listingRecord->listPC=address;				// change list output to show this address

			for(i=0;!fail&&i<numWords;i++)
			{
				temp=wordValues[i];
				if(temp&~currentProcessor->coreMask)	// see if word uses more bits than it should (if so, kill upper bits and generate warning)
				{
					AssemblyComplaint(NULL,false,"Program word too large. Truncated to core size. (%04X)\n",temp);
					temp&=currentProcessor->coreMask;
				}
				length=strlen(listingRecord->listObjectString);
				if(length+6<MAXSTRING)
				{
					sprintf(&listingRecord->listObjectString[length],"%04X ",temp);		// create list file output
				}
				outputBytes[0]=temp&0xFF;				// data is written as little endian words
				outputBytes[1]=temp>>8;

				fail=!AddBytesToSegment(currentSegment,address*2,outputBytes,2);	// each memory location is 16 bits wide
				address++;
			}
		}
	}
	else
	{
		AssemblyComplaint(NULL,true,"Attempt to write with no current segment\n");
	}
	return(!fail);
}

static bool GenerateWord(int wordValue,LISTING_RECORD *listingRecord)
// output the wordValue to the current segment
// This will return false only if a "hard" error occurs
{
	unsigned char
		outputBytes[2];
	int
		length;
	bool
		fail;

	fail=false;
	if(currentSegment)
	{
		if(currentSegment->currentPC<currentProcessor->ROMSize)
		{
			if(!intermediatePass)				// only do the real work if necessary
			{
				if(wordValue&~currentProcessor->coreMask)	// see if word uses more bits than it should (if so, kill upper bits and generate warning)
				{
					AssemblyComplaint(NULL,false,"Program word too large. Truncated to core size. (%04X)\n",wordValue);
					wordValue&=currentProcessor->coreMask;
				}
				length=strlen(listingRecord->listObjectString);
				if(length+6<MAXSTRING)
				{
					sprintf(&listingRecord->listObjectString[length],"%04X ",wordValue);		// create list file output
				}
				outputBytes[0]=wordValue&0xFF;	// data is written as little endian words
				outputBytes[1]=wordValue>>8;

				fail=!AddBytesToSegment(currentSegment,currentSegment->currentPC*2,outputBytes,2);	// each memory location is 16 bits wide
			}
			currentSegment->currentPC++;
		}
		else
		{
			AssemblyComplaint(NULL,true,"Attempt to write code outside address space\n");
		}
	}
	else
	{
		AssemblyComplaint(NULL,true,"Code cannot occur outside of a segment\n");
	}
	return(!fail);
}

static unsigned char
	heldByte;
static int
	byteGenCounter;

static void StartByteGeneration()
// some opcodes generate bytes at a time which need to be packed into words
// this makes that easier
{
	byteGenCounter=0;
}

static bool GenerateByte(unsigned char byteValue,LISTING_RECORD *listingRecord)
// output the byteValue to the current segment
{
	bool
		fail;

	fail=false;
	if(byteGenCounter&1)
	{
		fail=!GenerateWord((heldByte<<8)+byteValue,listingRecord);
	}
	else
	{
		heldByte=byteValue;					// just hold the byte until it is later flushed
	}
	byteGenCounter++;
	return(!fail);
}

static bool FlushByteGeneration(LISTING_RECORD *listingRecord)
// flush out any odd byte
{
	if(byteGenCounter&1)
	{
		return(GenerateByte(0,listingRecord));
	}
	return(true);
}

static bool ParseOperandSeparator(char *theLine,int *lineIndex)
// Expect a comma, step over one if found
{
	ParseWhiteSpace(theLine,lineIndex);		// move past any white space
	if(theLine[*lineIndex]==',')			// does this look like an opcode separator?
	{
		(*lineIndex)++;						// step over it
		return(true);
	}
	return(false);
}

static bool ParseRangeSeparator(char *theLine,int *lineIndex)
// Expect a dash, step over one if found
{
	ParseWhiteSpace(theLine,lineIndex);		// move past any white space
	if(theLine[*lineIndex]=='-')			// does this look like an opcode separator?
	{
		(*lineIndex)++;						// step over it
		return(true);
	}
	return(false);
}

static bool ParseDestination(char *theLine,int *lineIndex,int *destination,bool *unresolved,bool *unspecified)
// try to find a destination register
{
	*destination=1;							// default if unresolved, or unspecified
	*unspecified=true;
	if(ParseOperandSeparator(theLine,lineIndex))
	{
		*unspecified=false;
		return(ParseExpression(theLine,lineIndex,destination,unresolved));
	}
	return(true);
}

static bool ParseNextExpression(char *theLine,int *lineIndex,int *value,bool *unresolved)
// look for a separator, and another expression
{
	if(ParseOperandSeparator(theLine,lineIndex))
	{
		return(ParseExpression(theLine,lineIndex,value,unresolved));
	}
	return(false);
}

static void ReportUnspecifiedDestination()
// complain that the operands are bad for a given opcode
{
	AssemblyComplaint(NULL,false,"Unspecified destination -- defaults to register file\n");
}

static bool CheckRAMLocation(int theValue,bool generateMessage,bool isError)
// NOTE: because the PICS use a banked memory scheme, and because
// we want to allow programmers to be able to specify memory locations
// as absolutes, we only test to see if theValue is out of the range
// specified by MAXRAM, and BADRAM.
// This means that the value may be larger than will fit in the field,
// but we assume the programmer has done the correct bank selecting
// to make sure he is referencing the place he thinks he is.
{
	if((theValue>=0)&&(theValue<=maxRAM)&&(!(badRAMMap[theValue>>3]&(1<<(theValue&7)))))
	{
		return(true);
	}

	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Invalid RAM location specified\n");
	}
	return(false);
}

static bool CheckRegister3Range(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as a 3 bit register
// complain if not
{
	if(testBadRAM)
	{
		return(CheckRAMLocation(theValue,generateMessage,isError));
	}
	return(true);
}

static bool CheckRegister5Range(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as a 5 bit register
// complain if not
{
	if(testBadRAM)
	{
		return(CheckRAMLocation(theValue,generateMessage,isError));
	}
	return(true);
}

static bool CheckRegister7Range(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as a 7 bit register
// complain if not
{
	if(testBadRAM)
	{
		return(CheckRAMLocation(theValue,generateMessage,isError));
	}
	return(true);
}

static bool CheckRegister8Range(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as a 8 bit register
// complain if not
// NOTE: because the PICS use a banked memory scheme, and because
// we want to allow programmers to be able to specify memory locations
// as absolutes, we only test to see if theValue is out of the range
// specified by MAXRAM, and BADRAM.
// This means that the value may be larger than will fit in the field,
// but we assume the programmer has done the correct bank selecting
// to make sure he is referencing the place he thinks he is.
{
	if(testBadRAM)
	{
		return(CheckRAMLocation(theValue,generateMessage,isError));
	}
	return(true);
}

static bool CheckPortRange(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as a 5 bit port
// complain if not
{
	if(theValue>=0&&theValue<32)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Port value out of range\n");
	}
	return(false);
}

static bool CheckDestinationRange(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as a 1 bit destination
// complain if not
{
	if(theValue>=0&&theValue<2)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Destination value out of range\n");
	}
	return(false);
}

static bool CheckHighLowRange(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as a 1 bit highlow flag
// complain if not
{
	if(theValue>=0&&theValue<2)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"High/Low value out of range\n");
	}
	return(false);
}

static bool CheckIncrementRange(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as a 1 bit increment flag
// complain if not
{
	if(theValue>=0&&theValue<2)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Increment value out of range\n");
	}
	return(false);
}

// pseudo op handling

static bool HandleDB(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// declaring bytes of data, or packed strings
{
	int
		value;
	bool
		unresolved;
	char
		outputString[MAXSTRING];
	int
		stringLength;
	bool
		done,
		fail;
	int
		i;

	fail=false;
	done=false;
	StartByteGeneration();
	while(!done&&!fail)
	{
		if(ParseQuotedString(theLine,lineIndex,'"','"',outputString,&stringLength))
		{
			i=0;
			while(!fail&&i<stringLength)
			{
				fail=!GenerateByte(outputString[i],listingRecord);
				i++;
			}
		}	
		else if(ParseExpression(theLine,lineIndex,&value,&unresolved))
		{
			CheckByteRange(value,true,true);
			fail=!GenerateByte(value&0xFF,listingRecord);
		}
		else
		{
			ReportBadOperands();
			done=true;
		}
		if(!done&&!fail)				// look for separator, or comment
		{
			if(ParseOperandSeparator(theLine,lineIndex))
			{
			}
			else if(ParseComment(theLine,lineIndex))
			{
				done=true;
			}
			else
			{
				ReportBadOperands();
				done=true;
			}
		}
	}
	if(!fail)
	{
		fail=!FlushByteGeneration(listingRecord);
	}
	return(!fail);
}

static bool HandleDW(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// declaring words of data, or packed strings
{
	int
		value;
	bool
		unresolved;
	char
		outputString[MAXSTRING];
	int
		stringLength;
	bool
		done,
		fail;
	int
		i;

	fail=false;
	done=false;
	StartByteGeneration();
	while(!done&&!fail)
	{
		if(ParseQuotedString(theLine,lineIndex,'"','"',outputString,&stringLength))
		{
			i=0;
			while(!fail&&i<stringLength)
			{
				fail=!GenerateByte(outputString[i],listingRecord);
				i++;
			}
			if(!fail)
			{
				fail=!FlushByteGeneration(listingRecord);
			}
		}	
		else if(ParseExpression(theLine,lineIndex,&value,&unresolved))
		{
			CheckWordRange(value,true,true);
			if(GenerateByte(value>>8,listingRecord))
			{
				fail=!GenerateByte(value&0xFF,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			ReportBadOperands();
			done=true;
		}
		if(!done&&!fail)				// look for separator, or comment
		{
			if(ParseOperandSeparator(theLine,lineIndex))
			{
			}
			else if(ParseComment(theLine,lineIndex))
			{
				done=true;
			}
			else
			{
				ReportBadOperands();
				done=true;
			}
		}
	}
	return(!fail);
}

static bool HandleDT(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// make a table of retlw instructions with associated data
{
	int
		value;
	bool
		unresolved;
	unsigned char
		insnByte;
	char
		outputString[MAXSTRING];
	int
		stringLength;
	bool
		done,
		fail;
	int
		i;

	fail=false;
	done=false;

	switch(currentProcessor->family)
	{
		case PIC_16C5X:
			insnByte=0x08;
			break;
		case PIC_16CXX:
			insnByte=0x34;
			break;
		case PIC_17CXX:
		default:
			insnByte=0xB6;
			break;
	}

	StartByteGeneration();
	while(!done&&!fail)
	{
		if(ParseQuotedString(theLine,lineIndex,'"','"',outputString,&stringLength))
		{
			i=0;
			while(!fail&&i<stringLength)
			{
				if(GenerateByte(insnByte,listingRecord))		// output the retlw instruction
				{
					fail=!GenerateByte(outputString[i],listingRecord);
				}
				else
				{
					fail=true;
				}
				i++;
			}
		}	
		else if(ParseExpression(theLine,lineIndex,&value,&unresolved))
		{
			CheckByteRange(value,true,true);
			if(GenerateByte(insnByte,listingRecord))			// output the retlw instruction
			{
				fail=!GenerateByte(value&0xFF,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			ReportBadOperands();
			done=true;
		}
		if(!done&&!fail)				// look for separator, or comment
		{
			if(ParseOperandSeparator(theLine,lineIndex))
			{
			}
			else if(ParseComment(theLine,lineIndex))
			{
				done=true;
			}
			else
			{
				ReportBadOperands();
				done=true;
			}
		}
	}
	return(!fail);
}

static bool HandleDS(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// move the PC forward by the given amount
{
	int
		value;
	bool
		unresolved;

	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			if(currentSegment)
			{
				if(!unresolved)
				{
					currentSegment->currentPC+=value;
				}
			}
			else
			{
				AssemblyComplaint(NULL,true,"'%s' outside of segment\n",opcode->name);
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(true);
}

static bool HandleConfig(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// write the given configuration bits out
{
	int
		value;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			fail=!GenerateWordsAtAddress(currentProcessor->configAddress,1,&value,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleIDLocs(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// write the ID locations for the given processor
{
	int
		value;
	bool
		unresolved;
	int
		IDValues[4];
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			if(currentProcessor->IDLocAddress)	// does this processor support ID?
			{
				IDValues[0]=(value>>12)&0x0F;
				IDValues[1]=(value>>8)&0x0F;
				IDValues[2]=(value>>4)&0x0F;
				IDValues[3]=(value)&0x0F;
				fail=!GenerateWordsAtAddress(currentProcessor->IDLocAddress,4,IDValues,listingRecord);
			}
			else
			{
				AssemblyComplaint(NULL,true,"'%s' invalid for this processor\n",opcode->name);
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleMaxRAM(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// maximum RAM location is being specified
{
	int
		i;
	int
		value;
	bool
		unresolved;

	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			if(!unresolved)
			{
				if((value>=0)&&(value<MAX_BAD_RAM))
				{
					maxRAM=value;
					testBadRAM=true;
					for(i=0;i<(MAX_BAD_RAM>>3);i++)
					{
						badRAMMap[i]=0;		// mark all of RAM as being valid
					}
				}
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(true);
}

static bool HandleBadRAM(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// unusable RAM locations are being specified
// NOTE: because some bozo working on mpasm decided that the ranges should be separated with
// -'s, it is not possible to parse an expression for the range values (since
// expressions can and often do contain -'s)
{
	int
		startValue,
		endValue;
	int
		i;
	bool
		done,
		parseFail;

	parseFail=false;
	done=false;
	while(!parseFail&&!done)
	{
		if(ParseNumber(theLine,lineIndex,&startValue))
		{
			if(ParseRangeSeparator(theLine,lineIndex))
			{
				if(!ParseNumber(theLine,lineIndex,&endValue))
				{
					ReportBadOperands();
					parseFail=true;
				}
			}
			else
			{
				endValue=startValue;
			}
			if(!parseFail)	// have start and end of range
			{
				if(testBadRAM)
				{
					if((startValue>=0)&&(endValue>=startValue)&&(endValue<=maxRAM))
					{
						for(i=startValue;i<=endValue;i++)
						{
							badRAMMap[i>>3]|=(1<<(i&7));		// mark bad ram locations into bit table
						}
					}
					else
					{
						ReportBadOperands();
						parseFail=true;
					}
				}
				else
				{
					AssemblyComplaint(NULL,true,"'%s' before __MAXRAM\n",opcode->name);
					parseFail=true;
				}
			}
			if(!parseFail&&!ParseOperandSeparator(theLine,lineIndex))	// need to see separator to continue
			{
				done=true;
			}
		}
		else		// failed to get a number when one was expected
		{
			ReportBadOperands();
			parseFail=true;
		}
	}
	if(!parseFail)
	{
		if(!ParseComment(theLine,lineIndex))				// make sure we got to the end
		{
			ReportBadOperands();
		}
	}
	return(true);
}

static bool HandleBankSel(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// Generate bank selection code
{
	int
		value;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			switch(currentProcessor->family)
			{
				case PIC_16C5X:
					if(currentProcessor->bankSelBits)
					{
						if(GenerateWord(0x04A4|((value&0x20)<<3),listingRecord))		// generate BSF, or BCF on bit 5 of FSR
						{
							if(currentProcessor->bankSelBits>1)
							{
								fail=!GenerateWord(0x04C4|((value&0x40)<<2),listingRecord);	// generate BSF, or BCF on bit 6 of FSR
							}
						}
						else
						{
							fail=true;
						}
					}
					break;
				case PIC_16CXX:
					if(currentProcessor->bankSelBits)
					{
						if(GenerateWord(0x1283|((value&0x80)<<3),listingRecord))			// generate BSF, or BCF on bit 5 of STATUS
						{
							if(currentProcessor->bankSelBits>1)
							{
								fail=!GenerateWord(0x1303|((value&0x100)<<2),listingRecord);	// generate BSF, or BCF on bit 6 of STATUS
							}
						}
						else
						{
							fail=true;
						}
					}
					break;
				case PIC_17CXX:
					if((currentProcessor->bankSelBits)&&(value&0xFF)>=0x20)
					{
						fail=!GenerateWord(0xBA00|((value>>4)&0xF0),listingRecord);		// generate MOVLR
					}
					else
					{
						fail=!GenerateWord(0xB800|((value>>8)&0x0F),listingRecord);		// generate MOVLB
					}
					break;
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleBankISel(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// Generate indirect bank selection code
{
	int
		value;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			switch(currentProcessor->family)
			{
				case PIC_16C5X:
					break;
				case PIC_16CXX:
					if(currentProcessor->bankISelBits)
					{
						fail=!GenerateWord(0x1383|((value&0x100)<<2),listingRecord);		// generate BSF, or BCF on bit 7 of STATUS
					}
					break;
				case PIC_17CXX:
					if((currentProcessor->bankISelBits)&&(value&0xFF)>=0x20)
					{
						fail=!GenerateWord(0xBA00|((value>>4)&0xF0),listingRecord);		// generate MOVLR
					}
					else
					{
						fail=!GenerateWord(0xB800|((value>>8)&0x0F),listingRecord);		// generate MOVLB
					}
					break;
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandlePageSel(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// Generate page selection code
{
	int
		value;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			switch(currentProcessor->family)
			{
				case PIC_16C5X:
					if(currentProcessor->ROMSize>0x200)
					{
						if(GenerateWord(0x04A3|((value&0x200)>>1),listingRecord))				// generate BSF, or BCF on bit 5 of STATUS
						{
							if(currentProcessor->ROMSize>0x400)
							{
								fail=!GenerateWord(0x04C3|((value&0x400)>>2),listingRecord);	// generate BSF, or BCF on bit 6 of STATUS
							}
						}
						else
						{
							fail=true;
						}
					}
					break;
				case PIC_16CXX:
					if(currentProcessor->ROMSize>0x800)
					{
						if(GenerateWord(0x3000|((value>>8)&0xFF),listingRecord))		// generate MOVLW (high byte of address)
						{
							fail=!GenerateWord(0x008A,listingRecord);					// generate MOVWF PCLATH
						}
						else
						{
							fail=true;
						}
					}
					break;
				case PIC_17CXX:
					if(GenerateWord(0xB000|((value>>8)&0xFF),listingRecord))		// generate MOVLW (high byte of address)
					{
						fail=!GenerateWord(0x0103,listingRecord);					// generate MOVWF PCLATH
					}
					else
					{
						fail=true;
					}
					break;
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

// opcode handling

static bool HandleIMPLICIT(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle implicit opcodes
{
	bool
		fail;

	fail=false;
	if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
	{
		fail=!GenerateWord(opcode->baseOpcode,listingRecord);
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleLITERAL8(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle literal 8 bit opcodes
{
	int
		immediate;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&immediate,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			CheckByteRange(immediate,true,true);
			immediate&=0xFF;
			fail=!GenerateWord(opcode->baseOpcode|immediate,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleCALL8(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle call opcodes (8 bits of operand, the 9th bit must be clear or an error is generated)
{
	int
		immediate;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&immediate,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			if(immediate<0||immediate>=1024||(immediate&(1<<8)))
			{
				AssemblyComplaint(NULL,true,"Address (0x%04X) is out of range\n",immediate);
			}
			immediate&=0xFF;
			fail=!GenerateWord(opcode->baseOpcode|immediate,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleCALL11(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle call opcodes (11 bits of operand)
{
	int
		immediate;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&immediate,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			if(immediate<0||immediate>=2048)
			{
				AssemblyComplaint(NULL,true,"Address (0x%04X) is out of range\n",immediate);
			}
			immediate&=0x7FF;
			fail=!GenerateWord(opcode->baseOpcode|immediate,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleGOTO9(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle goto opcodes (9 bit operand)
{
	int
		immediate;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&immediate,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			if(immediate<0||immediate>=1024)
			{
				AssemblyComplaint(NULL,true,"Address (0x%04X) is out of range\n",immediate);
			}
			immediate&=0x1FF;
			fail=!GenerateWord(opcode->baseOpcode|immediate,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleGOTO11(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle goto opcodes (11 bit operand)
{
	int
		immediate;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&immediate,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			if(immediate<0||immediate>=2048)
			{
				AssemblyComplaint(NULL,true,"Address (0x%04X) is out of range\n",immediate);
			}
			immediate&=0x7FF;
			fail=!GenerateWord(opcode->baseOpcode|immediate,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleGOTO13(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle goto/call opcodes (13 bit operand)
{
	int
		immediate;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&immediate,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			if(immediate<0||immediate>=8192)
			{
				AssemblyComplaint(NULL,true,"Address (0x%04X) is out of range\n",immediate);
			}
			immediate&=0x1FFF;
			fail=!GenerateWord(opcode->baseOpcode|immediate,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER3(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with a 3 bit register
{
	int
		immediate;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&immediate,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			CheckRegister3Range(immediate,true,false);
			immediate&=0x07;
			fail=!GenerateWord(opcode->baseOpcode|immediate,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER5(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with a 5 bit register
{
	int
		theRegister;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&theRegister,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			CheckRegister5Range(theRegister,true,false);
			theRegister&=0x1F;
			fail=!GenerateWord(opcode->baseOpcode|theRegister,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER5_DEST(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with a 5 bit register and a destination
{
	int
		theRegister;
	bool
		registerUnresolved;
	int
		destination;
	bool
		destUnresolved,
		destUnspecified;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&theRegister,&registerUnresolved))		// get the register
	{
		if(ParseDestination(theLine,lineIndex,&destination,&destUnresolved,&destUnspecified))
		{
			if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
			{
				CheckRegister5Range(theRegister,true,false);
				CheckDestinationRange(destination,true,true);
				theRegister&=0x1F;
				destination&=0x01;
				fail=!GenerateWord(opcode->baseOpcode|(destination<<5)|theRegister,listingRecord);
				if(destUnspecified)
				{
					ReportUnspecifiedDestination();
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER5_BIT(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with a 5 bit register and a 3 bit bit index
{
	int
		theRegister;
	bool
		registerUnresolved;
	int
		bit;
	bool
		bitUnresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&theRegister,&registerUnresolved))
	{
		if(ParseNextExpression(theLine,lineIndex,&bit,&bitUnresolved))
		{
			if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
			{
				CheckRegister5Range(theRegister,true,false);
				Check8BitIndexRange(bit,true,true);
				theRegister&=0x1F;
				bit&=0x07;
				fail=!GenerateWord(opcode->baseOpcode|(bit<<5)|theRegister,listingRecord);
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER7(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with a 7 bit register
{
	int
		theRegister;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&theRegister,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			CheckRegister7Range(theRegister,true,false);
			theRegister&=0x7F;
			fail=!GenerateWord(opcode->baseOpcode|theRegister,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER7_DEST(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with a 7 bit register and a destination
{
	int
		theRegister;
	bool
		registerUnresolved;
	int
		destination;
	bool
		destUnresolved,
		destUnspecified;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&theRegister,&registerUnresolved))
	{
		if(ParseDestination(theLine,lineIndex,&destination,&destUnresolved,&destUnspecified))
		{
			if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
			{
				CheckRegister7Range(theRegister,true,false);
				CheckDestinationRange(destination,true,true);
				theRegister&=0x7F;
				destination&=0x01;
				fail=!GenerateWord(opcode->baseOpcode|(destination<<7)|theRegister,listingRecord);
				if(destUnspecified)
				{
					ReportUnspecifiedDestination();
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER7_BIT(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with a 7 bit register and a 3 bit bit index
{
	int
		theRegister;
	bool
		registerUnresolved;
	int
		bit;
	bool
		bitUnresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&theRegister,&registerUnresolved))
	{
		if(ParseNextExpression(theLine,lineIndex,&bit,&bitUnresolved))
		{
			if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
			{
				CheckRegister7Range(theRegister,true,false);
				Check8BitIndexRange(bit,true,true);
				theRegister&=0x7F;
				bit&=0x07;
				fail=!GenerateWord(opcode->baseOpcode|(bit<<7)|theRegister,listingRecord);
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER8(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with an 8 bit register
{
	int
		theRegister;
	bool
		unresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&theRegister,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			CheckRegister8Range(theRegister,true,false);
			theRegister&=0xFF;
			fail=!GenerateWord(opcode->baseOpcode|theRegister,listingRecord);
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER8_DEST(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with an 8 bit register and 1 bit destination
{
	int
		theRegister;
	bool
		registerUnresolved;
	int
		destination;
	bool
		destUnresolved,
		destUnspecified;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&theRegister,&registerUnresolved))		// get the register
	{
		if(ParseDestination(theLine,lineIndex,&destination,&destUnresolved,&destUnspecified))
		{
			if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
			{
				CheckRegister8Range(theRegister,true,false);
				CheckDestinationRange(destination,true,true);
				theRegister&=0xFF;
				destination&=0x01;
				fail=!GenerateWord(opcode->baseOpcode|(destination<<8)|theRegister,listingRecord);
				if(destUnspecified)
				{
					ReportUnspecifiedDestination();
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER8_PORT(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with an 8 bit register and 5 bit port
{
	int
		theRegister;
	bool
		registerUnresolved;
	int
		port;
	bool
		portUnresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&theRegister,&registerUnresolved))		// get the register
	{
		if(ParseNextExpression(theLine,lineIndex,&port,&portUnresolved))
		{
			if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
			{
				CheckRegister8Range(theRegister,true,false);
				CheckPortRange(port,true,true);
				theRegister&=0xFF;
				port&=0x1F;
				fail=!GenerateWord(opcode->baseOpcode|(port<<8)|theRegister,listingRecord);
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandlePORT_REGISTER8(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with a 5 bit port and 8 bit register
{
	int
		theRegister;
	bool
		registerUnresolved;
	int
		port;
	bool
		portUnresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&port,&portUnresolved))
	{
		if(ParseNextExpression(theLine,lineIndex,&theRegister,&registerUnresolved))
		{
			if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
			{
				CheckPortRange(port,true,true);
				CheckRegister8Range(theRegister,true,false);
				theRegister&=0xFF;
				port&=0x1F;
				fail=!GenerateWord(opcode->baseOpcode|(port<<8)|theRegister,listingRecord);
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleHIGHLOW_INCREMENT_REGISTER8(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with high/low bit, an increment bit, and an 8 bit register
{
	int
		theRegister;
	bool
		registerUnresolved;
	int
		highLow;
	bool
		highLowUnresolved;
	int
		increment;
	bool
		incrementUnresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&highLow,&highLowUnresolved))
	{
		if(ParseNextExpression(theLine,lineIndex,&increment,&incrementUnresolved))
		{
			if(ParseNextExpression(theLine,lineIndex,&theRegister,&registerUnresolved))
			{
				if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
				{
					CheckHighLowRange(highLow,true,true);
					CheckIncrementRange(increment,true,true);
					CheckRegister8Range(theRegister,true,false);
					theRegister&=0xFF;
					highLow&=0x01;
					increment&=0x01;
					fail=!GenerateWord(opcode->baseOpcode|(highLow<<9)|(increment<<8)|theRegister,listingRecord);
				}
				else
				{
					ReportBadOperands();
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleHIGHLOW_REGISTER8(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with high/low bit, and an 8 bit register
{
	int
		theRegister;
	bool
		registerUnresolved;
	int
		highLow;
	bool
		highLowUnresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&highLow,&highLowUnresolved))
	{
		if(ParseNextExpression(theLine,lineIndex,&theRegister,&registerUnresolved))
		{
			if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
			{
				CheckHighLowRange(highLow,true,true);
				CheckRegister8Range(theRegister,true,false);
				theRegister&=0xFF;
				highLow&=0x01;
				fail=!GenerateWord(opcode->baseOpcode|(highLow<<9)|theRegister,listingRecord);
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleREGISTER8_BIT(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes with an 8 bit register and a 3 bit bit index
{
	int
		theRegister;
	bool
		registerUnresolved;
	int
		bit;
	bool
		bitUnresolved;
	bool
		fail;

	fail=false;
	if(ParseExpression(theLine,lineIndex,&theRegister,&registerUnresolved))
	{
		if(ParseNextExpression(theLine,lineIndex,&bit,&bitUnresolved))
		{
			if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
			{
				CheckRegister8Range(theRegister,true,false);
				Check8BitIndexRange(bit,true,true);
				theRegister&=0xFF;
				bit&=0x07;
				fail=!GenerateWord(opcode->baseOpcode|(bit<<8)|theRegister,listingRecord);
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleOpcode(void *theOpcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// handle opcodes for this processor
{
	bool
		fail;
	OPCODE
		*actualOpcode;

	actualOpcode=(OPCODE *)theOpcode;
	fail=false;
	if(actualOpcode->pseudoHandler)
	{
		fail=!actualOpcode->pseudoHandler(actualOpcode,theLine,lineIndex,listingRecord);
	}
	else
	{
		switch(actualOpcode->type)
		{
			case OT_IMPLICIT:
				fail=!HandleIMPLICIT(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_LITERAL8:
				fail=!HandleLITERAL8(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_CALL8:
				fail=!HandleCALL8(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_CALL11:
				fail=!HandleCALL11(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_GOTO9:
				fail=!HandleGOTO9(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_GOTO11:
				fail=!HandleGOTO11(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_GOTO13:
				fail=!HandleGOTO13(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER3:
				fail=!HandleREGISTER3(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER5:
				fail=!HandleREGISTER5(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER5_DEST:
				fail=!HandleREGISTER5_DEST(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER5_BIT:
				fail=!HandleREGISTER5_BIT(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER7:
				fail=!HandleREGISTER7(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER7_DEST:
				fail=!HandleREGISTER7_DEST(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER7_BIT:
				fail=!HandleREGISTER7_BIT(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER8:
				fail=!HandleREGISTER8(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER8_DEST:
				fail=!HandleREGISTER8_DEST(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER8_PORT:
				fail=!HandleREGISTER8_PORT(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_PORT_REGISTER8:
				fail=!HandlePORT_REGISTER8(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_HIGHLOW_INCREMENT_REGISTER8:
				fail=!HandleHIGHLOW_INCREMENT_REGISTER8(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_HIGHLOW_REGISTER8:
				fail=!HandleHIGHLOW_REGISTER8(actualOpcode,theLine,lineIndex,listingRecord);
				break;
			case OT_REGISTER8_BIT:
				fail=!HandleREGISTER8_BIT(actualOpcode,theLine,lineIndex,listingRecord);
				break;
		}
	}
	return(!fail);
}

static void *MatchOpcode(char *theOpcode)
// match opcodes for the current processor, return NULL if none matched
{
	void
		*result;

	result=NULL;

	result=STFindDataForName(opcodeSymbolsPseudo,theOpcode);				// search pseudo opcodes first

	if(!result)
	{
		switch(currentProcessor->family)
		{
			case PIC_16C5X:
				result=STFindDataForName(opcodeSymbols16C5X,theOpcode);
				break;
			case PIC_16CXX:
				result=STFindDataForName(opcodeSymbols16CXX,theOpcode);
				break;
			case PIC_17CXX:
				result=STFindDataForName(opcodeSymbols17CXX,theOpcode);
				break;
		}
	}

	return(result);
}

static bool SelectProcessor(PROCESSOR *theProcessor)
// A processor in this family is being selected to assemble with
// NOTE: the PICs use a word based PC, so alter the PC to have word meaning
// when the processor is selected
{
	currentProcessor=(PIC_PROCESSOR *)theProcessor->processorData;	// remember which processor has been selected
	testBadRAM=false;												// selecting a processor kills the bad ram test
	if(currentSegment)
	{
		currentSegment->currentPC=(currentSegment->currentPC+1)/2;
		currentSegment->codeGenOffset=(currentSegment->codeGenOffset+1)/2;
	}
	return(true);
}

static void DeselectProcessor(PROCESSOR *theProcessor)
// A processor in this family is being deselected
{
	if(currentSegment)
	{
		currentSegment->currentPC*=2;
		currentSegment->codeGenOffset*=2;
	}
}

static void UnInitFamily()
// undo what InitFamily did
{
	STDisposeSymbolTable(opcodeSymbols17CXX);
	STDisposeSymbolTable(opcodeSymbols16CXX);
	STDisposeSymbolTable(opcodeSymbols16C5X);
	STDisposeSymbolTable(opcodeSymbolsPseudo);
}

static bool InitFamily()
// initialize symbol table
{
	unsigned int
		i;
	bool
		fail;

	testBadRAM=false;			// by default, do not test for bad RAM
	fail=false;
	if((opcodeSymbolsPseudo=STNewSymbolTable()))
	{
		for(i=0;!fail&&(i<sizeof(OpcodesPseudo)/sizeof(OPCODE));i++)
		{
			if(!STAddEntryAtEnd(opcodeSymbolsPseudo,OpcodesPseudo[i].name,&OpcodesPseudo[i]))
			{
				fail=true;
			}
		}
		if(!fail)
		{
			if((opcodeSymbols16C5X=STNewSymbolTable()))
			{
				for(i=0;!fail&&(i<sizeof(Opcodes16C5X)/sizeof(OPCODE));i++)
				{
					if(!STAddEntryAtEnd(opcodeSymbols16C5X,Opcodes16C5X[i].name,&Opcodes16C5X[i]))
					{
						fail=true;
					}
				}
				if(!fail)
				{
					if((opcodeSymbols16CXX=STNewSymbolTable()))
					{
						for(i=0;!fail&&(i<sizeof(Opcodes16CXX)/sizeof(OPCODE));i++)
						{
							if(!STAddEntryAtEnd(opcodeSymbols16CXX,Opcodes16CXX[i].name,&Opcodes16CXX[i]))
							{
								fail=true;
							}
						}
						if(!fail)
						{
							if((opcodeSymbols17CXX=STNewSymbolTable()))
							{
								for(i=0;!fail&&(i<sizeof(Opcodes17CXX)/sizeof(OPCODE));i++)
								{
									if(!STAddEntryAtEnd(opcodeSymbols17CXX,Opcodes17CXX[i].name,&Opcodes17CXX[i]))
									{
										fail=true;
									}
								}
								if(!fail)
								{
									return(true);
								}
								STDisposeSymbolTable(opcodeSymbols17CXX);
							}
						}
						STDisposeSymbolTable(opcodeSymbols16CXX);
					}
				}
				STDisposeSymbolTable(opcodeSymbols16C5X);
			}
		}
		STDisposeSymbolTable(opcodeSymbolsPseudo);
	}
	return(false);
}

// NOTE: there is some uncertainty in these values as I do not have
// the spec page for every device here.
// If you find a problem, email the author with upated values

static PIC_PROCESSOR
	pic12c508	={PIC_16C5X,0x0FFF,0,0,0x0200,0x0FFF,0x0200},
	pic12c509	={PIC_16C5X,0x0FFF,2,0,0x0400,0x0FFF,0x0400},
	pic12ce518	={PIC_16C5X,0x0FFF,0,0,0x0200,0x0FFF,0x0200},
	pic12ce519	={PIC_16C5X,0x0FFF,2,0,0x0400,0x0FFF,0x0400},
	pic12c671	={PIC_16CXX,0x3FFF,2,1,0x0400,0x2007,0x2000},
	pic12c672	={PIC_16CXX,0x3FFF,2,1,0x0800,0x2007,0x2000},
	pic16c505	={PIC_16C5X,0x0FFF,2,0,0x0400,0x0FFF,0x0400},
	pic16c52	={PIC_16C5X,0x0FFF,0,0,0x0180,0x0FFF,0x0180},
	pic16c54	={PIC_16C5X,0x0FFF,0,0,0x0200,0x0FFF,0x0200},
	pic16c55	={PIC_16C5X,0x0FFF,0,0,0x0200,0x0FFF,0x0200},
	pic16c56	={PIC_16C5X,0x0FFF,0,0,0x0400,0x0FFF,0x0400},
	pic16c57	={PIC_16C5X,0x0FFF,2,0,0x0800,0x0FFF,0x0800},
	pic16c58	={PIC_16C5X,0x0FFF,2,0,0x0800,0x0FFF,0x0800},

	pic14000	={PIC_16CXX,0x3FFF,1,0,0x1000,0x2007,0x2000},
	pic16c554	={PIC_16CXX,0x3FFF,1,0,0x0200,0x2007,0x2000},
	pic16c556	={PIC_16CXX,0x3FFF,1,0,0x0400,0x2007,0x2000},
	pic16c558	={PIC_16CXX,0x3FFF,1,0,0x0800,0x2007,0x2000},
	pic16c61	={PIC_16CXX,0x3FFF,1,0,0x0400,0x2007,0x2000},
	pic16c62	={PIC_16CXX,0x3FFF,1,0,0x0800,0x2007,0x2000},
	pic16c620	={PIC_16CXX,0x3FFF,1,0,0x0200,0x2007,0x2000},
	pic16c621	={PIC_16CXX,0x3FFF,1,0,0x0400,0x2007,0x2000},
	pic16c622	={PIC_16CXX,0x3FFF,1,0,0x0800,0x2007,0x2000},
	pic16c63	={PIC_16CXX,0x3FFF,1,0,0x1000,0x2007,0x2000},
	pic16c64	={PIC_16CXX,0x3FFF,1,0,0x0800,0x2007,0x2000},
	pic16c642	={PIC_16CXX,0x3FFF,1,0,0x1000,0x2007,0x2000},
	pic16c65	={PIC_16CXX,0x3FFF,1,0,0x1000,0x2007,0x2000},
	pic16c66	={PIC_16CXX,0x3FFF,2,1,0x2000,0x2007,0x2000},
	pic16c662	={PIC_16CXX,0x3FFF,1,0,0x1000,0x2007,0x2000},
	pic16c67	={PIC_16CXX,0x3FFF,2,1,0x2000,0x2007,0x2000},
	pic16c71	={PIC_16CXX,0x3FFF,1,0,0x0400,0x2007,0x2000},
	pic16c710	={PIC_16CXX,0x3FFF,1,0,0x0200,0x2007,0x2000},
	pic16c711	={PIC_16CXX,0x3FFF,1,0,0x0400,0x2007,0x2000},
	pic16c715	={PIC_16CXX,0x3FFF,1,0,0x0800,0x2007,0x2000},
	pic16c72	={PIC_16CXX,0x3FFF,1,0,0x0800,0x2007,0x2000},
	pic16c73	={PIC_16CXX,0x3FFF,1,0,0x1000,0x2007,0x2000},
	pic16c74	={PIC_16CXX,0x3FFF,1,0,0x1000,0x2007,0x2000},
	pic16c76	={PIC_16CXX,0x3FFF,2,1,0x2000,0x2007,0x2000},
	pic16c77	={PIC_16CXX,0x3FFF,2,1,0x2000,0x2007,0x2000},
	pic16f83	={PIC_16CXX,0x3FFF,1,0,0x0200,0x2007,0x2000},
	pic16c84	={PIC_16CXX,0x3FFF,1,0,0x0400,0x2007,0x2000},
	pic16f84	={PIC_16CXX,0x3FFF,1,0,0x0400,0x2007,0x2000},
	pic16f873	={PIC_16CXX,0x3FFF,2,1,0x1000,0x2007,0x2000},
	pic16f874	={PIC_16CXX,0x3FFF,2,1,0x1000,0x2007,0x2000},
	pic16f876	={PIC_16CXX,0x3FFF,2,1,0x2000,0x2007,0x2000},
	pic16f877	={PIC_16CXX,0x3FFF,2,1,0x2000,0x2007,0x2000},
	pic16c923	={PIC_16CXX,0x3FFF,2,1,0x1000,0x2007,0x2000},
	pic16c924	={PIC_16CXX,0x3FFF,2,1,0x1000,0x2007,0x2000},

	pic17c42	={PIC_17CXX,0xFFFF,0,0,0x0800,0xFE00,0x0000},
	pic17c43	={PIC_17CXX,0xFFFF,4,4,0x1000,0xFE00,0x0000},
	pic17c44	={PIC_17CXX,0xFFFF,4,4,0x2000,0xFE00,0x0000},
	pic17c52	={PIC_17CXX,0xFFFF,4,4,0x2000,0xFE00,0x0000},
	pic17c56	={PIC_17CXX,0xFFFF,4,4,0x4000,0xFE00,0x0000};

// processors handled here (the constuctors for these variables link them to the global
// list of processors that the assembler knows how to handle)

static PROCESSOR_FAMILY
	processorFamily("Microchip PIC",InitFamily,UnInitFamily,SelectProcessor,DeselectProcessor,MatchOpcode,HandleOpcode);

static PROCESSOR
// 16C5x
	processor00(&processorFamily,	"12c508"	,&pic12c508),
	processor01(&processorFamily,	"12c509"	,&pic12c509),
	processor02(&processorFamily,	"12ce518"	,&pic12ce518),
	processor03(&processorFamily,	"12ce519"	,&pic12ce519),
	processor04(&processorFamily,	"12c671"	,&pic12c671),
	processor05(&processorFamily,	"12c672"	,&pic12c672),
	processor06(&processorFamily,	"16c505"	,&pic16c505),
	processor07(&processorFamily,	"16c52"		,&pic16c52),
	processor08(&processorFamily,	"16c54"		,&pic16c54),
	processor09(&processorFamily,	"16c55"		,&pic16c55),
	processor10(&processorFamily,	"16c56"		,&pic16c56),
	processor11(&processorFamily,	"16c57"		,&pic16c57),
	processor12(&processorFamily,	"16c58"		,&pic16c58),
// 16Cxx
	processor13(&processorFamily,	"14000"		,&pic14000),
	processor14(&processorFamily,	"16c554"	,&pic16c554),
	processor15(&processorFamily,	"16c556"	,&pic16c556),
	processor16(&processorFamily,	"16c558"	,&pic16c558),
	processor17(&processorFamily,	"16c61"		,&pic16c61),
	processor18(&processorFamily,	"16c62"		,&pic16c62),
	processor19(&processorFamily,	"16c620"	,&pic16c620),
	processor20(&processorFamily,	"16c621"	,&pic16c621),
	processor21(&processorFamily,	"16c622"	,&pic16c622),
	processor22(&processorFamily,	"16c63"		,&pic16c63),
	processor23(&processorFamily,	"16c64"		,&pic16c64),
	processor24(&processorFamily,	"16c642"	,&pic16c642),
	processor25(&processorFamily,	"16c65"		,&pic16c65),
	processor26(&processorFamily,	"16c66"		,&pic16c66),
	processor27(&processorFamily,	"16c662"	,&pic16c662),
	processor28(&processorFamily,	"16c67"		,&pic16c67),
	processor29(&processorFamily,	"16c71"		,&pic16c71),
	processor30(&processorFamily,	"16c710"	,&pic16c710),
	processor31(&processorFamily,	"16c711"	,&pic16c711),
	processor32(&processorFamily,	"16c715"	,&pic16c715),
	processor33(&processorFamily,	"16c72"		,&pic16c72),
	processor34(&processorFamily,	"16c73"		,&pic16c73),
	processor35(&processorFamily,	"16c74"		,&pic16c74),
	processor36(&processorFamily,	"16c76"		,&pic16c76),
	processor37(&processorFamily,	"16c77"		,&pic16c77),
	processor38(&processorFamily,	"16f83"		,&pic16f83),
	processor39(&processorFamily,	"16c84"		,&pic16c84),
	processor40(&processorFamily,	"16f84"		,&pic16f84),
	processor41(&processorFamily,	"16f873"	,&pic16f873),
	processor42(&processorFamily,	"16f874"	,&pic16f874),
	processor43(&processorFamily,	"16f876"	,&pic16f876),
	processor44(&processorFamily,	"16f877"	,&pic16f877),
	processor45(&processorFamily,	"16c923"	,&pic16c923),
	processor46(&processorFamily,	"16c924"	,&pic16c924),
// 17CXX
	processor47(&processorFamily,	"17c42"		,&pic17c42),
	processor48(&processorFamily,	"17c43"		,&pic17c43),
	processor49(&processorFamily,	"17c44"		,&pic17c44),
	processor50(&processorFamily,	"17c52"		,&pic17c52),
	processor51(&processorFamily,	"17c56"		,&pic17c56);
